/*
 * NeuralController.java
 *
 * Created on 4 May 2007, 14:05
 *
 */

package ecu;

import au.edu.ecu.is.fuzzy.*;
import au.edu.ecu.is.neural.*;

import java.util.*;

import simplerace.*;

public class NeuralController implements Controller, Constants
{
    private FuzzyController fuzzy;
    private FuzzyWaypointCorrectionController fuzzyCorrection;
    private MultiLayerPerceptron mlp;
    private double speedLimiterFactor;
    private double distanceLimiterFactor;
    
    private static final double TARGET_SPEED = 5.0;
    private double diagonal = Math.sqrt(400*400 + 400*400);
    private static final int INPUTS = 2;
    private static final int HIDDEN = 5;
    private static final int OUTPUTS = 2;
    private double orientation;
    private double[][] weights =
    {
        {-0.762},
        {0.907},
        {-0.761, -0.075, 0.079},
        {-0.578, -0.485, -1.916},
        {-0.561, 0.261, -0.755},
        {1.25, 0.419, -1.407},
        {-0.534, -1.086, -0.083},
        {0.12, -0.141, -0.91, -0.074, 0.042, -0.756},
        {-0.271, -0.982, 0.523, -0.516, 1.101, -0.335}
    };
    
    public NeuralController()
    {
        try
        {
            fuzzy = new FuzzyController();
            fuzzyCorrection = new FuzzyWaypointCorrectionController();
            speedLimiterFactor = 1.0;
            distanceLimiterFactor = 1.0;
            
            int[] layerSizes = new int[]{INPUTS, HIDDEN, OUTPUTS};
            
            mlp = new MultiLayerPerceptron(layerSizes, new Random(), 0, 0, true);
            mlp.setWeights(weights);
        }
        catch (Exception e)
        {
        }
    }
    
    public void reset()
    {
    }
    
    public double getSpeedToNextWaypoint(Vector2d velocity, double orientation)
    {
        double magnitude = velocity.mag();
        double movementDir = Math.atan2(velocity.x, velocity.y);
        double dirDifference =  movementDir - orientation;
        return Math.cos(dirDifference) * magnitude;
    }
    
    public double getDistanceToNextWaypoint(Vector2d waypoint, Vector2d position)
    {
        //System.out.println(position.dist(waypoint));
        
        
        return position.dist(waypoint) / diagonal;
    }
    
    private static boolean errorReported = false;
    
    private double normalize(double actualValue, double minimum, double maximum)
    {
        return ((actualValue - minimum)/(maximum - minimum));
    }
    
    public double getAngleToWaypoint(Vector2d pos, double orientation, Vector2d wp)
    {
        
        double xDiff = wp.x - pos.x;
        double yDiff = wp.y - pos.y;
        double angle = Math.atan2(yDiff, xDiff);
        angle = orientation - angle;
        return Utils.correctAngle(angle);
    }
    
    public int control(SensorModel inputs)
    {
        double speed = Math.abs(inputs.getSpeed());
        double distance = inputs.getDistanceToNextWaypoint() * distanceLimiterFactor;
        double enemySpeed = getSpeedToNextWaypoint(inputs.getOtherVehicleVelocity(), inputs.getOtherVehicleOrientation());
        double enemyDistance = getDistanceToNextWaypoint(inputs.getNextWaypointPosition(),inputs.getOtherVehiclePosition());
        double enemyAngle = getAngleToWaypoint(inputs.getOtherVehiclePosition(),inputs.getOtherVehicleOrientation(),
                inputs.getNextWaypointPosition());
        
        try
        {
            double newAngle, newDistance;
            double newSpeed = inputs.getSpeed();
            newAngle = inputs.getAngleToNextWaypoint();
            newDistance = inputs.getDistanceToNextWaypoint();
            double newNextAngle = inputs.getAngleToNextNextWaypoint();
            
            if(inputs.otherVehicleIsPresent())
            {
                mlp.setInputs(new double[]{speed / enemySpeed, distance / enemyDistance});
                if (mlp.getOutput(0) > mlp.getOutput((1)))
                {
                    
                    newAngle = inputs.getAngleToNextNextWaypoint();
                    newDistance = inputs.getDistanceToNextNextWaypoint();
                }
                else
                {
                    newAngle = inputs.getAngleToNextWaypoint();
                    newDistance = inputs.getDistanceToNextWaypoint();
                }
                
                if ((fuzzyCorrection.control(enemyDistance, enemyAngle) > 0))
                {
                    newAngle = inputs.getAngleToNextWaypoint();
                    newDistance = inputs.getDistanceToNextWaypoint();
                }
            }
            else
            {
                newAngle = inputs.getAngleToNextWaypoint();
                newDistance = inputs.getDistanceToNextWaypoint();
            }
            
            return fuzzy.control(newAngle,newDistance,newSpeed,newNextAngle,speedLimiterFactor);
        }
        catch(Exception e)
        {
            // default to heuristic controller if there is an unexpected error
            if(!errorReported)
            {
                e.printStackTrace(System.err);
                errorReported = true;
            }
            int steering = inputs.getAngleToNextWaypoint() > 0 ? 0 : 2;
            int driving = inputs.getSpeed() > TARGET_SPEED ? 1 : 2;
            return driving * 3 + steering;
        }
    }
    
    public void setSpeedLimiterFactor(double speedLimiterFactor)
    {
        this.speedLimiterFactor = speedLimiterFactor;
    }
}
