/*
 * FuzzyController.java
 *
 * Created on 8 April 2007, 08:35
 *
 */

package ecu;

import au.edu.ecu.is.fuzzy.*;

import java.util.Vector;

import simplerace.*;

public class FuzzyController implements Controller, Constants
{
    private SugenoRuleSet controlRules;
    private SugenoRuleSet speedRules;
    private SugenoRuleSet driftingControlRules;
    private FuzzyVariable steeringCorrectionX;
    private FuzzyVariable speedError;
    private FuzzyVariable steeringError;
    private FuzzyVariable targetSpeed;
    private FuzzyVariable distanceToWayPoint;
    private FuzzyVariable angleOfWaypoint;
    private FuzzyVariable nosePositioning;
    private FuzzyVariable speedOfCar;
    private FuzzyVariable IPositionX;
    private FuzzyVariable angleBetweenWaypoints;
    private static final double TARGET_SPEED = 4.0;
    private static final double MAX_SPEED_ERROR = 30.0;
    private static final double MAX_DISTANCE = 2.0;
    
    public FuzzyController()
    {
        try
        {
            controlRules = new SugenoRuleSet();
            speedRules = new SugenoRuleSet();
            driftingControlRules = new SugenoRuleSet();
            steeringCorrectionX = new FuzzyVariable("steering correction x", "setting", -1, 1, 1);
            steeringError = new FuzzyVariable("steering error", "radians", -Math.PI, Math.PI, 2);
            FuzzySet left = new FuzzySet("left", -Math.PI, -Math.PI, -Math.PI/2, 0.0);
            FuzzySet ok = new FuzzySet("straight", -Math.PI/10, 0.0, 0.0, Math.PI/10);
            FuzzySet right = new FuzzySet("right", 0.0, Math.PI/2, Math.PI, Math.PI);
            
            steeringError.add(left);
            steeringError.add(ok);
            steeringError.add(right);
            FuzzySet[] steeringErrorSets = {left, ok, right};
            
            speedError = new FuzzyVariable("speed error", "kliks", -MAX_SPEED_ERROR, MAX_SPEED_ERROR, 2);
            FuzzySet slow = new FuzzySet("slow", -MAX_SPEED_ERROR, -MAX_SPEED_ERROR, -2.0, 0.0);
            FuzzySet steady = new FuzzySet("steady", -0.2, 0, 0, 0.2);
            FuzzySet fast = new FuzzySet("fast", 0.0, 2.0, MAX_SPEED_ERROR, MAX_SPEED_ERROR);
            speedError.add(slow);
            speedError.add(steady);
            speedError.add(fast);
            FuzzySet[] speedErrorSets = {slow, steady, fast};
            
            targetSpeed = new FuzzyVariable("target speed","kliks", -20, 20,2);
            
            distanceToWayPoint = new FuzzyVariable("distance to waypoint","km",0, 2,3);
            
            FuzzySet veryNear = new FuzzySet("very near", 0, 0, 0.02, 0.025);
            FuzzySet near = new FuzzySet("near", 0.02, 0.06, 0.08, 0.1);
            FuzzySet medium = new FuzzySet("medium", 0.03, 0.083, 0.159, 0.394);
            FuzzySet far = new FuzzySet("far", 0.28, 0.283, 0.374, 0.394);
            FuzzySet veryFar = new FuzzySet("very far", 0.373, 0.635, 2.0, 2.0);
            distanceToWayPoint.add(veryNear);
            distanceToWayPoint.add(near);
            distanceToWayPoint.add(medium);
            distanceToWayPoint.add(far);
            distanceToWayPoint.add(veryFar);
            
            FuzzySet[] distanceSets = {veryNear, near, medium, far, veryFar};
            
            nosePositioning = new FuzzyVariable("nose position","choice", 0, 2, 2);
            speedOfCar = new FuzzyVariable("car speed", "klicks", 0, 20, 2);
            FuzzySet verySlowSpeed = new FuzzySet("very slow", 0, 0, 1, 2);
            FuzzySet slowSpeed = new FuzzySet("slow", 1.5, 3, 4, 5);
            FuzzySet mediumSpeed = new FuzzySet("medium", 4.5, 6, 7, 8.5);
            FuzzySet fastSpeed = new FuzzySet("fast", 8, 10, 12, 13.5);
            FuzzySet veryFastSpeed = new FuzzySet("very fast", 13, 15, 20, 20);
            
            speedOfCar.add(verySlowSpeed);
            speedOfCar.add(slowSpeed);
            speedOfCar.add(mediumSpeed);
            speedOfCar.add(fastSpeed);
            speedOfCar.add(veryFastSpeed);
            
            // drifting logic 2
//            IPositionX = new FuzzyVariable("Ouside", "m", -30, 30, 2);
            
            angleBetweenWaypoints = new FuzzyVariable("Apex location", "radians", -Math.PI, Math.PI, 2);
            FuzzySet behindLeft = new FuzzySet("behindLeft", -Math.PI, -Math.PI, -Math.PI, -2.51);
            FuzzySet hardLeft = new FuzzySet("Hard left", -2.99, -2.51, -2.09, -1.57);
            FuzzySet leftDir     = new FuzzySet("Left",-1.75, -1.57, -0.79, 0);
            FuzzySet straightAhead = new FuzzySet("Straight", -0.39, 0, 0, 0.39);
            FuzzySet rightDir    = new FuzzySet("Right", 0, 0.79, 1.57, 1.75);
            FuzzySet hardRight = new FuzzySet("Hard right", 1.75, 2.2, 2.34, 2.55);
            FuzzySet behindRight    = new FuzzySet("behind Right", 2.07, Math.PI, Math.PI, Math.PI);
            angleBetweenWaypoints.add(behindLeft);
            angleBetweenWaypoints.add(hardLeft);
            angleBetweenWaypoints.add(leftDir);
            angleBetweenWaypoints.add(straightAhead);
            angleBetweenWaypoints.add(rightDir);
            angleBetweenWaypoints.add(hardRight);
            angleBetweenWaypoints.add(behindRight);
            
            FuzzySet[] angleSets = {behindLeft, hardLeft, leftDir, straightAhead, rightDir, hardRight, behindRight};
            FuzzySet[] speedingSets = {verySlowSpeed, slowSpeed, mediumSpeed, fastSpeed, veryFastSpeed};
            
            double[][] settingsX =
            {
                {1, 1, 1},
                {0, 0, 0},
                {-1, -1, -1}
            };
            
            double [][] settingsY =
            {
                {-0.15, -20, -20, -10, -9.95},
                {16.72, -13.94, -11.46, -8, -20},
                {-20, 2.31, 2, 4.41, 16.22},
                {-3.85, 16.69, 7 ,10, 20},
                {15.86,2, 2.43, 4.15, 8.82},
                {-20, 14.34, -9.42, 2.01, -20},
                {-11,-20,-7,-12.16,-20}
            };
            
            double [][]steeringSettings =
            {
                {1, -1, -1},
                {1, 0, -1},
                {1, 1, -1}
            };
            
            double [][]driftingSettings =
            {
                {0, 0, 0, 0, 0},
                {0, 0, 1, 1, 1},
                {0, 0, 0, 0, 0},
                {0, 0, 0, 0, 0},
                {0, 0, 0, 0, 0}
            };
            
            controlRules.addRuleMatrix(
                    steeringError, steeringErrorSets,
                    speedError, speedErrorSets,
                    steeringCorrectionX,
                    settingsX
                    );
            
            speedRules.addRuleMatrix(
                    angleBetweenWaypoints, angleSets,
                    distanceToWayPoint, distanceSets,
                    targetSpeed,
                    settingsY);
            
            driftingControlRules.addRuleMatrix(
                    distanceToWayPoint, distanceSets,
                    speedOfCar, speedingSets,
                    nosePositioning, driftingSettings);
        }
        catch(FuzzyException e)
        {
            throw new RuntimeException(e);
        }
    }
    
    // ******************************************************************************
    
    // This is a utility method that converts
    //
    //      * dx, a desired change of speed in the x direction (side to side), and
    //      * dy, a desired change of speed in the y direction (forwards or backwards)
    //
    //  to
    //
    //      a control action (an integer between 0 and 9)
    
    private static final double SMALL = 0.025;
    private static final int controls[] = {4, 7, 1, -1, 5, 8, 2, -1, 3, 6, 0};
    private static final int width = 400;
    private static final int height = 400;
    private static final double diagonal = Math.sqrt(width*width + height*height);
    
    private int getControl(double dx, double dy)
    {
        // -1 <= dx <= 1
        // -1 <= dy <= 1
        
        int code = (((dx < -SMALL)?1:0)<<3) |
                (((dx > SMALL)?1:0)<<2) |
                (((dy < -SMALL)?1:0)<<1)|
                ((dy > SMALL)?1:0);
        
        return controls[code];
    }
    // *********************************************************************************
    
    public void reset()
    {
    }
    
    private static boolean errorReported = false;
    
    public double getAngleToWaypoint(Vector2d pos, double orientation, Vector2d wp)
    {
        
        double xDiff = wp.x - pos.x;
        double yDiff = wp.y - pos.y;
        double angle = Math.atan2(yDiff, xDiff);
        angle = orientation - angle;
        return Utils.correctAngle(angle);
    }
    
    public int control(SensorModel inputs)
    {
        return control(inputs.getAngleToNextWaypoint(), inputs.getDistanceToNextWaypoint(), inputs.getSpeed(), inputs.getAngleToNextNextWaypoint(),1.0);
    }
    
    public int control(double angle, double distFromWaypoint, double speed, double nextAngle, double speedLimiterFactor)
    {
        double currentWaypointDistance = distFromWaypoint;
        double currentAngleToWayPoint = angle;
        
        try
        {
            currentWaypointDistance = Math.min(currentWaypointDistance, MAX_DISTANCE);
            distanceToWayPoint.setValue(currentWaypointDistance);
            speedOfCar.setValue(Math.abs(speed));
            driftingControlRules.update();
            if (nosePositioning.getValue() > 0.4)
            {
                currentAngleToWayPoint = nextAngle;
            }
            
            steeringError.setValue(currentAngleToWayPoint);
            angleBetweenWaypoints.setValue(currentAngleToWayPoint);
            speedRules.update();
            
            double crispSpeedError = Math.max(-MAX_SPEED_ERROR, Math.min(MAX_SPEED_ERROR, speed-targetSpeed.getValue()));
            speedError.setValue(crispSpeedError);
            
            controlRules.update();
            
            double x = steeringCorrectionX.getValue();
            
            double modifiedTargetSpeed = targetSpeed.getValue() * speedLimiterFactor;
            double y = speed > modifiedTargetSpeed? -1:1;
            
            return getControl(x, y);
        }
        catch(Exception e)
        {
            // default to heuristic controller if there is an unexpected error
            if(!errorReported)
            {
                e.printStackTrace(System.err);
                errorReported = true;
            }
            int steering = angle > 0 ? 0 : 2;
            int driving = speed > TARGET_SPEED ? 1 : 2;
            return driving * 3 + steering;
        }
    }
}
