/*
 * FuzzyWaypointCorrectionController.java
 *
 * Created on 2 July 2007, 14:50
 *
 */

package ecu;

import au.edu.ecu.is.fuzzy.*;

import java.util.Vector;

import simplerace.*;

public class FuzzyWaypointCorrectionController implements Constants
{
    private SugenoRuleSet orbitCorrectionRules;
    private FuzzyVariable distanceToWaypoint;
    private FuzzyVariable angleToWaypoint;
    private FuzzyVariable actualWaypointToGo;
    
    public FuzzyWaypointCorrectionController()
    {
        try
        {
            actualWaypointToGo = new FuzzyVariable("Choice","units",0, 1, 2);
            orbitCorrectionRules = new SugenoRuleSet();
            distanceToWaypoint = new FuzzyVariable("distance to waypoint","km",0, 2,3);
            
            // these distance values has not been optimized.
            FuzzySet veryNear = new FuzzySet("very near", 0, 0, 0.02, 0.025);
            FuzzySet near = new FuzzySet("near", 0.02, 0.06, 0.08, 0.1);
            FuzzySet medium = new FuzzySet("medium", 0.075, 0.15, 0.2, 0.25);
            FuzzySet far = new FuzzySet("far", 0.23, 0.3, 0.4, 0.5);
            FuzzySet veryFar = new FuzzySet("very far", 0.45, 0.7, 2.0, 2.0);
            distanceToWaypoint.add(veryNear);
            distanceToWaypoint.add(near);
            distanceToWaypoint.add(medium);
            distanceToWaypoint.add(far);
            distanceToWaypoint.add(veryFar);
            
            FuzzySet[] distanceSets = {veryNear, near, medium, far, veryFar};
            
            // used in speed logic
            angleToWaypoint = new FuzzyVariable("Apex location", "radians", -Math.PI, Math.PI, 2);
            FuzzySet behindLeft = new FuzzySet("behindLeft", -Math.PI, -Math.PI, -Math.PI, -Math.PI/1.25);
            FuzzySet hardLeft = new FuzzySet("Hard left", -Math.PI/1.05, -Math.PI/1.25, -Math.PI/1.5, -Math.PI/2);
            FuzzySet leftDir     = new FuzzySet("Left",-Math.PI/1.8, -Math.PI/2, -Math.PI/4, 0);
            FuzzySet straightAhead = new FuzzySet("Straignt", -Math.PI/8, 0, 0, Math.PI/8);
            FuzzySet rightDir    = new FuzzySet("Right", 0, Math.PI/4, Math.PI/2, Math.PI/1.8);
            FuzzySet hardRight = new FuzzySet("Hard right", Math.PI/2, Math.PI/1.5, Math.PI/1.25, Math.PI/1.05);
            FuzzySet behindRight    = new FuzzySet("behind Right", Math.PI/1.25, Math.PI, Math.PI, Math.PI);
            
            angleToWaypoint.add(behindLeft);
            angleToWaypoint.add(hardLeft);
            angleToWaypoint.add(leftDir);
            angleToWaypoint.add(straightAhead);
            angleToWaypoint.add(rightDir);
            angleToWaypoint.add(hardRight);
            angleToWaypoint.add(behindRight);
            FuzzySet[] angleSets = {behindLeft, hardLeft, leftDir, straightAhead, rightDir, hardRight, behindRight};
            
            // now we have the rules
            double[][] settings =
            {
                // veryNear, near, medium, far, veryFar
                {0, 0, 0, 0, 0},// behind left
                {0, 0, 0, 0, 0}, // hard left
                {0, 0, 0, 0, 0}, // left
                {1, 1, 1 ,1, 1}, // straightAhead
                {0,0, 0, 0, 0},   // right
                {0, 0, 0, 0, 0}, // hard Right
                {0,0,0,0,0} // behind right
            };
            
            orbitCorrectionRules.addRuleMatrix(
                    angleToWaypoint, angleSets, distanceToWaypoint, distanceSets, actualWaypointToGo, settings);
        }
        catch (FuzzyException e)
        {
            
            throw new RuntimeException(e);
        }
        
    }
    
    public double getAngleToWaypoint(Vector2d pos, double orientation, Vector2d wp)
    {
        
        double xDiff = wp.x - pos.x;
        double yDiff = wp.y - pos.y;
        double angle = Math.atan2(yDiff, xDiff);
        angle = orientation - angle;
        return Utils.correctAngle(angle);
    }
    
    public int control(double distance, double angle)
    {
        try
        {
            angleToWaypoint.setValue(angle);
            distanceToWaypoint.setValue(distance);
            
            orbitCorrectionRules.update();
            if (actualWaypointToGo.getValue() > 0.5)
                return 0;
            else
                return 1;
        }
        catch (FuzzyException e)
        {
            throw new RuntimeException(e);
        }
    }        
}
